<?php

namespace Drupal\cms_content_sync\Controller;

use Drupal\cms_content_sync\Entity\Flow;
use Drupal\cms_content_sync\Entity\Pool;
use Drupal\cms_content_sync\PullIntent;
use Drupal\cms_content_sync\SyncCore\Exception\TimeoutException;
use Drupal\Core\Url;
use Drupal\Core\Controller\ControllerBase;

/**
 * Pull controller.
 */
class FlowPull extends ControllerBase {

  /**
   * Pull all entities of this Flow.
   *
   * @param $cms_content_sync_flow
   * @param $pull_mode
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse|null
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function pull($cms_content_sync_flow, $pull_mode) {
    /**
     * @var \Drupal\cms_content_sync\Entity\Flow $flow
     */
    $flow = \Drupal::entityTypeManager()
      ->getStorage('cms_content_sync_flow')
      ->load($cms_content_sync_flow);

    $force_pull = FALSE;
    if ($pull_mode == 'all_entities') {
      $force_pull = TRUE;
    }

    $result = FlowPull::pullAll($flow, $force_pull);

    $operations = [];

    foreach ($result as $id => $operation) {
      $operations[] = [
        '\Drupal\cms_content_sync\Controller\FlowPull::batch',
        [$operation],
      ];
    }

    $batch = [
      'title' => t('Pull all'),
      'operations' => $operations,
      'finished' => '\Drupal\cms_content_sync\Controller\FlowPull::batchFinished',
    ];
    batch_set($batch);

    return batch_process(Url::fromRoute('entity.cms_content_sync_flow.collection'));
  }

  /**
   * Kindly ask the Sync Core to pull all entities that are auto pulled.
   *
   * @param \Drupal\cms_content_sync\Entity\Flow $flow
   * @param bool $force
   *
   * @return \Drupal\cms_content_sync\SyncCore\Interfaces\Syndication\IPullAll[]
   */
  public static function pullAll($flow, $force = FALSE) {
    $flow_import = $force ? [PullIntent::PULL_AUTOMATICALLY] : [PullIntent::PULL_AUTOMATICALLY, PullIntent::PULL_MANUALLY];
    $pool_import = $force ? [Pool::POOL_USAGE_FORCE] : [Pool::POOL_USAGE_FORCE, Pool::POOL_USAGE_ALLOW];

    /**
     * @var \Drupal\cms_content_sync\SyncCore\Interfaces\Syndication\IPullAll[] $result
     */
    $result = [];

    $pools = Pool::getAll();

    foreach ($flow->getEntityTypeConfig() as $id => $type) {
      $entity_type_name = $type['entity_type_name'];
      $bundle_name = $type['bundle_name'];

      if ($type['handler'] == Flow::HANDLER_IGNORE) {
        continue;
      }

      if (!in_array($type['import'], $flow_import)) {
        continue;
      }

      $entity_type_pools = [];
      if (isset($type['import_pools'])) {
        foreach ($type['import_pools'] as $pool_id => $state) {
          if (!isset($entity_type_pools[$pool_id])) {
            $entity_type_pools[$pool_id] = [];
          }
          $entity_type_pools[$pool_id]['import'] = $state;
        }
      }

      foreach ($entity_type_pools as $pool_id => $definition) {
        if (empty($pools[$pool_id])) {
          continue;
        }
        $pool = $pools[$pool_id];
        $import = $definition['import'] ?? NULL;

        if (!in_array($import, $pool_import)) {
          continue;
        }

        $result[] = $pool
          ->getClient()
          ->getSyndicationService()
          ->pullAll($flow->id, $entity_type_name, $bundle_name)
          ->fromPool($pool->id)
          ->force($force);
      }
    }

    return $result;
  }

  /**
   * Batch pull finished callback.
   *
   * @param $success
   * @param $results
   * @param $operations
   */
  public static function batchFinished($success, $results, $operations) {
    $failed = 0;
    $empty = 0;
    $synchronized = 0;
    foreach ($results as $result) {
      if ($result['type'] == 'FAILURE') {
        $failed++;
      }
      elseif ($result['type'] == 'EMPTY') {
        $empty++;
      }
      else {
        $synchronized += $result['total'];
      }
    }

    if ($failed) {
      \Drupal::messenger()->addMessage(t('Failed to pull from %failed entity pools.', ['%failed' => $failed]));
    }
    if ($empty) {
      \Drupal::messenger()->addMessage(t('%empty entity pools were empty or had no new entities.', ['%empty' => $empty]));
    }
    if ($synchronized) {
      \Drupal::messenger()->addMessage(t('%synchronized entities have been pulled.', ['%synchronized' => $synchronized]));
    }
  }

  /**
   * Batch pull callback for the pull-all operation.
   *
   * @param \Drupal\cms_content_sync\SyncCore\Interfaces\Syndication\IPullAll $operation
   * @param array $context
   *
   * @throws \Drupal\cms_content_sync\SyncCore\Exception\SyncCoreException
   */
  public static function batch($operation, &$context) {
    if (empty($context['sandbox']['operation'])) {
      $context['sandbox']['operation'] = $operation->execute();

      if (!$operation->total()) {
        $context['results'][] = [
          'type' => 'EMPTY',
        ];
        return;
      }
    }

    /**
     * @var \Drupal\cms_content_sync\SyncCore\Interfaces\Syndication\IPullAll $operation
     */
    $operation = $context['sandbox']['operation'];

    $total = $operation->total();

    try {
      $progress = $operation->progress();

      if ($progress < $total) {
        // Don't spam the Sync Core...
        sleep(5);
      }

      if ($progress == $total) {
        $context['results'][] = ['type' => 'success', 'total' => $total];
      }

      $context['finished'] = $progress / $operation->total();
      $context['message'] = 'Pulled ' . $progress . ' of ' . $operation->total() . ' ' . $operation->getTypeMachineName() . '.' . $operation->getBundleMachineName() . ' from ' . $operation->getPoolMachineName() . '...';
    }
    // Ignore timeouts, just wait until the Sync Core becomes responsive again.
    catch (TimeoutException $e) {
    }
  }

}
